unit HashUnit;

interface
type

  THashObject=class(TObject)
    s:string;
    t:TObject;
    deleted:boolean;
  end;
  THashStr = class(TObject)
    Objects:array  of THashObject;
    maxhash, orighashsize:integer;
    maxloadfactor:single;
    maxused:integer; {based on maxloading factor from create}
    count:integer;  {number of table positions occupied}
    firstentry:integer; {index of first entry}
    {Fields just to keep stats for testing}
    statsOn:boolean;
    totsearchlength:integer;
    maxsearchlength:integer; {longest collision set}
    nbrsearches:integer;
    constructor create(newmaxhash:integer; newmaxloading:single;keepstats:boolean);
    destructor destroy;    override;
    function find(s:string;var index:integer):boolean;
    procedure delete(index:integer);
    function AddObject(s:string;t:Tobject):boolean;
    function indexof(s:string):integer;
    function getfirst:integer;
    procedure resethash;
    procedure rehash;
    function  hash(s:string):integer;
    procedure clear;
    procedure ClearAndFreeObj;
  end;

implementation
  uses sysutils, classes, dialogs;

 {Thash Methods}

  {***************** THash.Create ******************}
  constructor THashStr.create(newmaxhash:integer; newmaxloading:single; keepstats:boolean);
  {Create THash object}
  begin
    inherited create;
    maxhash:=newmaxHash;
    orighashsize:=newmaxhash;
    setlength(Objects,maxhash);
    maxloadfactor:=newmaxloading;
    StatsOn:=keepstats;
    resethash;
  end;

  (*
  function THashStr.hash(s:string):integer;
  {OK for random stings - terrible for permutations because
    every key hashes to the same number}
  var
    i:integer;
  begin
    result:=1;
    for i:=1 to length(s) do result:=(result * ord(s[i])) mod maxhash;
  end;

  *)

  function THashStr.hash(s:string):integer;
   var
    i:integer;
  begin
    result:=1;
    for i:=1 to length(s) do result:=((result shl 4)+ ord(s[i])) mod maxhash{length(Objects)};
  end;




  destructor THashStr.destroy;
  {destroy it - probably not necessary}
  begin
    setlength(Objects,0);
    inherited;
  end;

  procedure THashStr.resethash;
  var
    i:integer;
  begin
    {zero out the array of hashed values}
    setlength(Objects,maxhash);
    for i:= 0 to length(Objects)-1 do Objects[i]:=nil;
    totsearchlength:=0;
    maxsearchlength:=0;
    nbrsearches:=0;
    count:=0;
    maxused:= trunc(maxloadfactor*maxhash);
  end;

 function THashStr.find(s:string;var index:integer):boolean;
  begin
    index:= indexof(s);
    result:=index>=0;
  end;

function THashStr.indexof(s:string):integer;
var
  foundindex:integer;

    function scansuccess(start,stopper:integer; var found:boolean):boolean;
      var
        i:integer;
      begin
        found:=false;
        result:=false;
        i:=start;
        while (i < stopper) and (Objects[i]<>nil)
            and (comparetext(Objects[i].s,s)<>0) do inc(i);
        {statistics keeping}
        If statson then
        begin
          totsearchlength:=totsearchlength+(i-start);
          inc(nbrsearches);
          if (i-start)>maxsearchlength then maxsearchlength:=i-start;
        end;
        If i<stopper then
        begin
          if (Objects[i]=nil) or (objects[i].deleted) then  found:=false
          else if (Comparetext(Objects[i].s,s)=0) then
          begin
            foundindex:=i;
            found:=true;
          end
          else result:=false;
        end;
      end;

  var
    h:integer;
    found:boolean;
  begin
    result:=-1;
    found:=false; foundindex:=-1;
    h:=hash(s);
    if h>maxhash then exit;
    if not scansuccess(h,maxhash{length(Objects)},found)
    then {loop around and try from beginning} scansuccess(0,h,found);
    result:=foundindex;
  end;


  function THashStr.GetFirst:integer;
  var
    i:integer;
  begin
    result:=-1;
    for i:= firstentry to high(Objects) do
    if (Objects[i]<>nil) and (not objects[i].deleted) then
    begin
      result:=i;
      firstentry:=i;
      break;
    end;
  end;

 procedure THashStr.delete(index:integer);
  begin
    objects[index].deleted:=true;
    objects[index].t:=nil;
    objects[index].s:='';
    dec(count);
  end;


 procedure THashStr.clear;
  var
    i:integer;
  begin
    for i:=low(objects) to high(objects) do
    if objects[i]<>nil then
    begin
      objects[i].free;
      objects[i]:=nil;
    end;
    count:=0;
    firstentry:=high(firstentry);
    setlength(objects,orighashsize);
  end;


 procedure THashStr.clearAndFreeObj;
  var
    i:integer;
  begin
    for i:=low(objects) to high(objects) do
    if objects[i]<>nil then
    begin
      if assigned(objects[i].t) then objects[i].t.free;
      objects[i].free;
      objects[i]:=nil;
    end;
    count:=0;
    firstentry:=high(firstentry);
    setlength(objects,orighashsize);
  end;


  {***************** AddObject *********************}
  function THashStr.AddObject(s:string; t:TObject):boolean;

      function scansuccess(start,stopper:integer; var found:boolean):boolean;
      var
        i:integer;
      begin
        result:=true;
        found:=false;
        i:=start;
        while (i < stopper) and (Objects[i]<>nil)
           and (not objects[i].deleted)
           {and (comparetext(Objects[i].s,s)<>0)} do inc(i);
        {statistics keeping}
        if statsOn then
        begin
          totsearchlength:=totsearchlength+(i-start);
          inc(nbrsearches);
          if (i-start)>maxsearchlength then maxsearchlength:=i-start;
        end;
        If i<stopper then
        begin
          if (Objects[i]=nil) or (objects[i].deleted) then
          begin
            If objects[i]=nil then Objects[i]:=THashObject.create;
            Objects[i].s:=s;
            Objects[i].t:=t;
            Objects[i].deleted:=false;
            found:=true;
            inc(count);
            if i<firstentry then firstentry:=i;
            if count>maxused then  rehash;
          end
          else if comparetext(Objects[i].s,s)=0
          {we are trying to add another one with the same key - let's flunk it}
          then result:=false;
        end
        else result:=false;
     end;

  var
    h:integer;
    found:boolean;
  begin
    if t=nil
      then showmessage('Nil object added');
    h:=hash(s);
    if not scansuccess(h,maxhash{length(Objects)},found) then {loop around and try from beginning}
    If not scansuccess(0,h,found) then
    begin
      {Could make this a warning and expand hash table}
      {To do that, we'd have to copy over existing data to temp array,
       expand the hash table with new maxhash, then rehash all numbers into the
       expanded table and make a recursive call to AddObjects for the current N}
      showmessage('Hash failed - table is full, contact vendor');
      found:=false;
    end;
    result:=found;
  end;

  procedure thashstr.rehash;
  {Max load factor exceeded, expand and rehash everything}
  var
    list:TStringlist;
    i:integer;
  begin
    list:=Tstringlist.create;
    for i:= 0 to length(Objects)-1 do
    if (Objects[i]<>nil) and (not objects[i].deleted)
     then list.addobject(Objects[i].s,Objects[i].t);
    if list.count<>count
    then showmessage('Invalid count on rehash, contactvendor');
    maxhash:=maxhash+orighashsize;
    setlength(Objects,maxhash);
    resethash;
    for i:=0 to list.count-1
    do addObject(list[i],list.objects[i]);
    list.free;
  end;


end.
