unit U_Puzzle15C;

{Copyright 2000,2006 Gary Darby, Intellitech Systems Inc., www.DelphiForFun.org

 This program may be used or modified for any non-commercial purpose
 so long as this original notice remains in place.
 All other rights are reserved
 }

{The 15 puzzle - scramble the sliders and put them back in sequence}
{Version 3.0 - Uses one of 2 methods:
        AStar algorithm with Manhattan heuristic (distance to goal =
        sum of column and row differences from goal for all tiles
        Improvements
          . Linear Conflict - two tiles in the column or row where
            they belong but revrsed - add 2 moves since that is
            the only way they can pass each other
          . Last Moves - in the target corner, last move must be from
            the corner that will end up empty, if tile to the left of
            final empty corner is not in the last column and the tile above
            the final empty corner is not in the bottom row, theManahttan distance
            will under estimate by at least 2 moves.
          . Corner Tiles - if a tile adjoining a corner is in place , but the
            corner tile is not the right one, at least 2 additional moves
            will be required to get it there.

}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, CheckLst, ComCtrls, HashUnit, ShellAPI;

type
  TBoardrec=record
     SliderIndex:byte;  {index of slider}
  end;

  TBoardRecs = array of byte {array of  slider indices};

  TStatsrec=record
     startstr:string;
     targetStr:string;
     Distance:integer;
     FVal:integer;
     Count:integer;
     countActive ,maxcounta, maxcountchecked, maxlevel:integer;
     starttime,stoptime:TDateTime;
     moves:integer;
     path:string;  {solution moves}
  end;


   {********************* TMoveObj definition ********************}
  TmoveObj=class(TObject)
  {This is the object used for tracking moves during solution search}
     public
     board:TBoardrecs;

     movetype:char; {the move that got us here: 'L','R','U','D'}
     prevMove:TMoveObj; {the previous configuration (parent)}
     movescore:integer; {distance from solution}
     cumscore:integer;  {sum of distances from prior moves}
     Fval:integer; {movescore+cumscore}
     empty:byte; {index of empty slot - not required - just for speed}
     level:integer;  {current depth of search}
     constructor create(parentmove:TMoveobj);
     function makeconfigkey:string;
     function BuildPath:string;
   end;


  TSliderBoard=class(TPanel)
  {The board}
    public
    Sidex,Sidey:integer; {width & height in tiles}
    Board: TBoardrecs; {slots for the sliders}
    target:TBoardrecs; {target tile nbrs indexed by slot nbr}
    targettype:integer; {0=std, 1=zigzag, 2= spiral}
    Targetparity:integer;
    TByTile:TBoardrecs; {target slots indexed by tile nbr}
    Sliders:array of TPanel; {the sliders}
    replaymoves:array of integer;  {array of tile to click to reshow solution}
    replayindex:integer;
    {the column and row of each slot - for speed, not really required}
    cols,rows:array of integer;
    empty :byte;{index of the empty slot - for speed, not really required}
    slotw:integer; {width & height of slider}
    UpdateMode:boolean;  {true=don't show animated moves}
    solving:boolean;  {Auto-solving is active}
    movecount:integer;  {count for display}
    weight:integer; {For experimenting with weighted A* , values of 1 to 100, 3 recommended}
    statsrec:TStatsrec;    {STATS}
    Heuristics: array[0..2] of boolean;
    Playmode:integer; {0=play, 1=setup}
    clicksOK:Boolean;  {Normally true, false during replay mode}
    {fields used in setup mode during drag operation}
    draggingOn:boolean;
    offset:TPoint;
    dragPosition:integer;
    dragstart:TPoint;

    {lists of TMoveObj used for auto solving}
    {Active lists - open items on 2 lists so that we can retrieve by fval score
     and by key- FVal list contains key for access by key in keylist}
    ActiveHashByKey:THashStr;
    ActiveListByFVal:TStringList;
    {Checked list - close items already checked}
    CheckedHash:THashStr;

    {A couple of callback events to communicate with caller for form updating}
    OnMove: TNotifyEvent;
    OnStatsRec: TNotifyEvent;

    constructor create(panel:TPanel; newSideX,newSideY:integer); reintroduce;
    destructor destroy; reintroduce;
    procedure resetboard;
    procedure move(index:integer); {move index slider}
    procedure slideit(sender:Tobject); {slider onclick event handler}
    procedure Startdrag(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure DragIt(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure Dropit(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    function  canmove(index:integer):boolean; {check if move posible}
    function  SolveIt(newtarget:TBoardrecs):boolean; {AStar for given target}
    function  makeTargetBoard1(var empty:byte):TBoardRecs;
    procedure setscore(MoveObj:TMoveObj;oldslot,newslot:integer);
    function findposition(n:byte):byte; {return slot index from tile nbr}
    procedure Followpath(s:string); {make moves specified in s string}
    procedure SetHeuristics(LC,LM,CT:boolean);
    procedure beginupdate; {stop move animation}
    procedure endupdate;   {restart move animation}
    function  solved:boolean;
    function parity(const testboard:TBoardRecs):integer;
    function Makeboardstr(const s:TBoardrecs):string;
  end;

  {******************* the form **********************}
  TForm1 = class(TForm)
    fstoptim: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    Label1: TLabel;
    Panel1: TPanel;
    Memo1: TMemo;
    MoveEdt: TEdit;
    SolveBtn: TButton;
    CancelBtn: TButton;
    BtnGroup: TGroupBox;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    SizeGrp: TRadioGroup;
    ResetBtn: TButton;
    ScrambleBtn: TButton;
    HeuristicsBox: TCheckListBox;
    ScrambleEdt: TEdit;
    UpDown1: TUpDown;
    WeightEdt: TEdit;
    TargetRgrp: TRadioGroup;
    PlayModeGrp: TRadioGroup;
    ClearStatsBtn: TButton;
    StaticText1: TStaticText;
    RichEdit1: TRichEdit;
    SolvedPanel: TPanel;
    ReplayBtn: TButton;
    StartOverBtn: TButton;
    procedure FormCreate(Sender: TObject);
    procedure ScrambleBtnClick(Sender: TObject);
    procedure SizeGrpClick(Sender: TObject);
    procedure SolveBtnClick(Sender: TObject);
    procedure CancelBtnClick(Sender: TObject);
    procedure ResetBtnClick(Sender: TObject);
    procedure HeuristicsBoxClick(Sender: TObject);
    procedure TargetRgrpClick(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure StaticText1Click(Sender: TObject);
    procedure StartOverBtnClick(Sender: TObject);
    procedure PlayModeGrpClick(Sender: TObject);
    procedure ClearStatsBtnClick(Sender: TObject);
    procedure ReplayBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
     SliderBoard:TSliderBoard;
     SavedBoard: TBoardrecs; {save the board here before changing}
     visited:TStringlist;
     sidex,sidey:integer;
     cheatcount, PCount:integer;
     CurStatStart:integer; {The line where the current case stats rec starts in display}
     CurStatSize:integer;
     //replaymoves:array[1..25] of integer; {array of tiles number to click to replat xolution}
     //replayindex:integer;  {the latest replay move}
     {event exits}
     Procedure UpdateMoveDisplay(sender:TObject);
     procedure Showstats(sender:TObject);
     procedure SaveBoardrecs;

  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

uses math;
const
  steps=20; {nbr of steps while moving a slider}
  scramblesteps=10; {Nbr of exchanges when Scramble button is pushed}
  //DefaultRandomSeed=12345;  {used to make repeatable boards}

{********************* TSliderBoard.Create *************************}
constructor TSliderBoard.create(panel:TPanel;newsidex,newsidey:integer);
{Create s new sliderboard}
var
  i:integer;
begin
  sidex:=newsidex;
  sidey:=newsidey;
  setlength(sliders,sidex*sidey);
  setlength(board,sidex*sidey);
  setlength(cols,sidex*sidey);
  setlength(rows,sidex*sidey);

  inherited create(panel.owner);
  slotw:=panel.width div sidex;
  height:=sidey*slotw;
  width:=sidex*slotw;
  left:=panel.Left;
  top:=panel.Top;
  parent:=panel.parent;
  color:=panel.color;
  font:=panel.font;

  for i:= low(board) to high(board) do
  begin
    board[i]:=i+1;
    cols[i]:=i mod sidex;
    rows[i]:=i div sidex;
  end;
  for i:=1 to sidex*sidey-1 do
  begin
    sliders[i]:=TPanel.create(self);
    with sliders[i] do
    begin
      parent:=self;
      left:= cols[i-1]*slotw;
      top:=  rows[i-1]*slotw;
      width:=slotw;
      height:=slotw;
      caption:=inttostr(i);
      tag:=i;
      bevelinner:=bvraised;
      bevelouter:=bvraised;
      bevelwidth:=4;
      onclick:=slideit;
      onMousedown:=startdrag;
      onmousemove:=dragit;
      onmouseup:=dropit;
    end;
  end;
  targettype:=0;
  resetboard;
  ActiveListByFVal:=TStringList.create;
  ActivelistByFval.sorted:=true;
  ActiveHashByKey:=THashStr.create(99991,0.50,false{true=keepstats});
  CheckedHash:=THashStr.create(99991,0.50,false);
  empty:=sidex*sidey-1;
  board[empty]:=0;
  cols[empty]:=empty mod sidex;
  rows[empty]:=empty div sidex;
  movecount:=0;
  solving:=false;
  invalidate;
  updatemode:=false;
  weight:=5;
  draggingon:=false;
  playmode:=0;
  for i:=low(heuristics) to high(heuristics) do heuristics[i]:=false;
end;

{***************** TsliderBoard.Destroy ***************}
destructor TSliderBoard.destroy;
  var
  i:integer;
  begin
    for i:= 1 to high(sliders) do sliders[i].free;
    setlength(board,0);
    setlength(sliders,0);
    inherited;
  end;

{******************* TSliderboard.ResetBoard ***************}
procedure Tsliderboard.resetboard;
var
  i:integer;
  begin
    replayindex:=0;
    board:=makeTargetBoard1(empty);
    setlength(target,length(board));
    target:=copy(board,low(board),length(board));
    for i:= low(board) to high(board) do
    if i<>empty then with sliders[board[i]] do
    begin
      left:=cols[i]*slotw;
      top:=rows[i]*slotw;
    end;
    color:=clGreen;
  end;


{********************* TSliderBoard.SlideIt **************}
procedure TSliderBoard.slideit(sender:TObject);
{Onclick exit for slider - move to empty if adjacent}
var
  i,j:integer;
begin
  {don't honor clicks while solving or doing a manual setup or replay}
  if solving or (playmode<>0) then exit;
  i:=TPanel(sender).tag;
  for j:=0 to high(board) do {find the location of the clicked tile}
  if board[j]=i then break;
  if canmove(j) then move(j);
  if Solved then
  begin
    showmessage('You did it!');
  end;
end;

{*********************** TSliderBoard.Move *****************}
procedure TSliderBoard.move(index:integer);
{move slider at board[index] to empty slot}
var
  stepx,stepy:integer;
  destx,desty,i:integer;
begin
  destx:=cols[empty]*slotw;
  desty:=rows[empty]*slotw;
  inc(MoveCount);
  if assigned(onMove) then OnMove(self);
  with sliders[board[index]] do
  begin
    stepx:=(destx-left) div steps;
    stepy:=(desty-top) div steps;
    if not updatemode then
    for i:= 1 to steps do
    begin
      left:=left+stepx;
      top:=top+stepy;
      invalidate;
      sleep(1);
      application.processmessages;
    end;
    left:=destx;
    top:=desty;
    invalidate;
  end;
  board[empty]:=board[index];
  board[index]:=0;
  empty:=index;
end;

{******************** TSliderBoard.CanMove *****************}
function TSliderBoard.canmove(index:Integer):boolean;
{check to make sure the empty slot is adjacent to index slot}
var
  x,y, xe,ye:integer;
begin
  x:=cols[index];
  y:=rows[index];
  xe:=cols[empty];
  ye:=rows[empty];
  result:=false;
  if x=xe then
    if (y+1=ye)
      or (y-1=ye)
    then result:=true
    else
  else
  if y=ye  then
    if (x+1=xe)
    or (x-1=xe)
    then result:=true;
end;

{****************** TSliderBoard.Setheuristics *************}
procedure TSliderBoard.SetHeuristics(LC,LM,CT:boolean);
{ See Korf "Finding Optimal Solutons to the Twenty-Four Puzzle" for
  expalantion of these}
begin
  Heuristics[2]:=LC; {Linear Conflict}
  Heuristics[1]:=LM; {Last Moves}
  Heuristics[0]:=CT; {Corner Tiles}
end;

{**************** TSliderBoard.MakeTargetBoard *************}
 function TSliderBoard.makeTargetBoard1(var empty:byte):TBoardRecs;
 {make a  target board}
  var
    i,j, index:integer;
    dirx,diry:integer;
    c,r:integer;
    loopcount, lim1, hlim, vlim :integer;
    tilenbr:integer;
  begin
   setlength(result,sidex*sidey);
  {if user wants a spiral on a 2 row board, give him a zigzag}
  if (sidey<3) and (targettype=2) then targettype:=1;
  case targettype of
  0:  {fill a standard target board, empty in bottom right}
    begin
      for i:= 0 to high(result)-1 do
      result[i]:=i+1;
      empty:=high(result);
      result[empty]:=0;
    end;
    1:  {zigzag}
      begin
        {fill a zigzag pattern}
        for i:= 0 to sidey-1 do
        for j:=0 to sidex-1 do
          if i mod 2 =1 then result[i*sidex+j]:=i*sidex+(sidex-j)
          else result[i*sidex+j]:=i*sidex+j+1;

        {put empty on the left side if even nbr of rows}
        if sidey mod 2 =0 then empty:=high(result)-sidex+1
        else empty:=high(result);
        result[empty]:=0;
      end;

    2: {spiral}
      begin
        tilenbr:=0;
        dirx:=0; diry:=-1;
        loopcount:=0;
        r:=0;
        c:=-1;
        hlim:=sidex-1; vlim:=sidey-1;
        while tilenbr<sidex*sidey do
        begin
          {tilenbr=0 then begin hlim:=sidex-1; vlim:=sidey-1;end
          else} if loopcount mod 2 =1 then begin dec(hlim); dec(vlim); end;
          if dirx<>0 then
          begin  diry:=dirx;  dirx:=0; lim1:=vlim; end
          else
          begin dirx:=-diry; diry:=0; lim1:=hlim; end;
          inc(loopcount);
          for i:=0 to lim1 do
          begin
            inc(tilenbr);
            r:=r+diry;
            c:=c+dirx;
            index:=r*sidex+c;
            If (tilenbr=sidex*sidey) then
            begin
              empty:=index;
              result[empty]:=0
            end
            else result[index]:=tilenbr;
          end;
        end;
      end; {spiral}
  end; {case}
  targetparity:=parity(result);
  setlength(TByTile,length(result)); {target array w/positions indexed by tile nbr}
  for i:=low(result) to high(result) do If result[i]<>0
  then TByTile[result[i]]:=i else tByTile[result[i]]:=255;
end;

{***************** TSliderBoard.FollowPath ************}
 procedure TSliderBoard.Followpath(s:string);
 {Given a string of moves, do it}
 var
   i,n:integer;
 begin
    n:=empty;
    setlength(replaymoves,length(s)+1);
    statsrec.path:='';
    for i:=1 to length(s) do
    begin
      case s[i] of
        'U': n:=n-sidex;
        'D': n:=n+sidex;
        'L': n:=n-1;
        'R': n:=n+1;
       end;
       replaymoves[i]:=n {board[n]}; {build array of moves for later replay}
       if board[n]<>0
       then
       begin
         with statsrec do path:=path+inttostr(board[n])+', ';
         move(n);
       end;
    end;
    with statsrec do delete(path,length(path),2);
 end;

 {*************TSliderBoard.SetScore **********************}
 procedure TSliderBoard.setscore(MoveObj:TMoveObj;oldslot,newslot:integer);
   var
     i,j,k,n:integer;
     coldiff, rowdiff, addscore:integer;

       {.......... FindNextNode ..................}
       function findNextNodePosition(n:byte):byte;
       {given a tile number, find the position index}
       var
         i:integer;
       begin
         with MoveObj do
         begin
           result:=0;
           for i:=low(board) to high(board) do
           if board[i]=n then
           begin
             result:=i;
             break;
           end;
         end;
       end; {findNextNode}


     function fullscorecalc:integer;
     {calc score from scratch - normally first time only}
     var
       i,j:integer;
     begin
       result:=0;
       with moveobj do
       for i:= low(board) to high(board) do
       if board[i]<>0 then   {no need to check empty}
       begin
         j:=TByTile[board[i]];
         rowdiff:=abs(rows[i]-rows[j]);
         coldiff:=abs(cols[i]-cols[j]);
         result:=result + (rowdiff + coldiff);
       end;
     end;


   begin {Setscore}
     with MoveObj do
     begin
       if prevmove=nil
       then {if no parent} movescore:=fullscorecalc
       else  {just calculate an incremental score from previous}
       begin
         movescore:=prevmove.movescore;
         i:=TByTile[moveobj.board[oldslot]];
         if (target[board[oldslot]]<255) and (i<255) then
         begin
           j:=oldslot;
           movescore:=movescore+(abs(rows[i]-rows[j])+abs(cols[i]-cols[j]));
           j:=newslot;
           movescore:=movescore-(abs(rows[i]-rows[j])+abs(cols[i]-cols[j]));
         end;
       end;
       addscore:=0;
       {handle additional heuristics}
       If heuristics[0] and (sidex>3) then    {Corner tiles}
       with moveobj do
       {handle only for target with empty in bottom right for now,
        check the other 3 corners}
       If (movescore>0) and (target[high(target)]=0) then
       begin
         {check top left}
         if (moveobj.board[0]<>0) and (moveobj.board[0]<>target[0]) then
         begin
           if  (moveobj.board[1]<>0) and  (moveobj.board[1]=target[1])
           then addscore:=addscore+2; {movescore:=movescore+2;}
           if (moveobj.board[sidex]<>0) and (moveobj.board[sidex]=target[sidex])
           then addscore:=addscore+2;{movescore:=movescore+2;}
         end;
         {check top right}
         if (moveobj.board[sidex-1]<>0) and (moveobj.board[sidex-1]<>target[sidex-1]) then
         begin
           if (moveobj.board[sidex-2]<>0) and(moveobj.board[sidex-2]=target[sidex-2])
           then addscore:=addscore+2;{movescore:=movescore+2;}
           if (moveobj.board[2*sidex-1]<>0) and(moveobj.board[2*sidex-1]=target[2*sidex-1])
           then addscore:=addscore+2;{movescore:=movescore+2;}
         end;

         {and finally, bottom left}
         j:=high(board)-sidex+1;
         if (moveobj.board[j]<>0) and (moveobj.board[j]<>target[j]) then
         begin
           i:=j-sidex;
           if (moveobj.board[i]<>0) and(moveobj.board[i]=target[i])
           then addscore:=addscore+2;
           i:=j+1;
           if (moveobj.board[i]<>0) and(moveobj.board[i]=target[i])
           then addscore:=addscore+2;
         end;
       end;



       If heuristics[1] then  {last  move heuristic}
       with moveobj do
       if (movescore>0) and (target[high(target)]=0) then {handle only empty in last position for now}
       begin
         {if tile to left of final empty is not in the last column
         and the tile above final empty is not in the bottom row,
         then we can add  2 moves}
         k:=target[high(target)-1];
         i:=findNextNodePosition(k); {this is where it is}
         k:=target[high(target)-sidex];
         j:=findNextNodePosition(k); {this is where it is}
         if  (rows[i]<>sidex-1) and (cols[j]<> sidey-1)
         then addscore:=addscore+2; {movescore:=movescore+2;}
       end;
       If heuristics[2] then   {linear conflicts}
       with moveobj do
       begin
         for i:=low(board) to high(board) do
         if board[i]<>0 then
         begin
           j:=TByTile[moveobj.board[i]]; {get target position}
           if (j<=high(board)) and (rows[i]=rows[j]) then{it's in the correct row}
           begin
             for k:=sidex*(i div sidex) to i-1 do {for leftside tiles in the row}
             if board[k]<>0 then {if it's not empty}
             begin
               n:=tByTile[moveobj.board[k]]; {get target pos for tiles left of j}
               {add 2 moves if it's in the correct row but on the wrong side}
               if (n<=high(board)) and (rows[n]=rows[i]) and (n>i)
               then addscore:=addscore+2;
             end
           end;
           if (cols[i]=cols[j]) then{it's in the correct column}
           begin
             k:=i mod sidex;  {get column top}
             while k<= i-sidex do {for everything above i}
             begin
               if moveobj.board[k]>0 then {if its not empty}
               begin
                 n:=tByTile[moveobj.board[k]]; {get target pos for tiles above j}
                 {if target is in the same column but below i, then add 2 moves}
                 if (n<=high(board)) and (cols[n]=cols[i]) and (n>i)
                 then addscore:=addscore+2;
               end;
               k:=k+sidex; {next in column}
             end
           end;
         end;
       end;
       fVal:=weight*(movescore+addscore) + cumscore;
       cumscore:=cumscore+movescore+addscore;
     end;
   end;


{****************** TSliderBoard.FindPosition ***************}
   function TSliderboard.findposition(n:byte):byte;
   {given a tile number, find its position index}
   var
     i:integer;
   begin
     result:=0;
     for i:=low(board) to high(board) do
     if board[i]=n then
     begin
       result:=i;
       break;
     end;
   end;



{******************** TSliderBoard.Solveit ******************}
function TSliderBoard.SolveIt(newtarget:TBoardrecs):boolean;
var
  index,i:integer;
  s:string;
  temp:TMoveObj;

  function makefvalkey(node:TMoveobj):string;
  begin
    result:=format('%6.6d',[node.fval])+node.makeconfigkey;
  end;


  procedure makesuccessors(parent:TMoveObj);

    procedure makemove(oldslot,newslot:integer; direction:char);
    var
       i:integer;
       nextnode,temp3:TMoveobj;
      begin {makemove}
        case direction of {don't move right back the way we came}
          'U': if parent.movetype='D' then exit;
          'D': if parent.movetype='U' then exit;
          'L': if parent.movetype='R' then exit;
          'R': if parent.movetype='L' then exit;
        end;
        nextnode:=TMoveObj.create(parent);
        nextnode.movetype:=direction;
        {swap the slider indices between empty and left slot}
        nextnode.board[oldslot]:=parent.board[newslot];
        nextnode.board[newslot]:=parent.board[oldslot];
        nextnode.empty:=newslot;
        setscore(nextnode,oldslot,newslot); {calculate the score stuff}
        {if same key is on the closed list with a large fval.
         we can delete it - it will not be used}
        if checkedHash.find(nextnode.makeconfigkey,index)
        and (TMoveObj(checkedHash.objects[index].t).fval>nextnode.fval)
        then
        begin
          TMoveobj(checkedHash.objects[index].t).free;
          checkedHash.Delete(index);
        end;
        {If it's on the active list with a larger Fval, we can delete
        from there too}
        If activeHashByKey.find(nextnode.makeconfigkey,index) then
        begin {there's aleady this config on the active list}
          temp3:=TMoveObj(ActiveHashBykey.Objects[index].t);
          if temp3.fval> nextnode.fval  {and the old has a higher fval}
          then {delete it}
          begin
            ActiveHashByKey.delete(index);
            i:=ActiveListByFVal.indexof(makefvalKey(temp3));
            if i>=0 then ActiveListByFval.delete(i)
            else showmessage('Check this 2');
            temp3.free;
            If not ActiveHashByKey.addobject(nextnode.makeconfigkey,nextnode)
            then showmessage('Add to ActiveByKey failed #1');
            ActiveListByFVal.addobject(makefvalkey(nextnode),nextnode);
          end
        end
        else
        begin  {in any event add it to the active list}
          if not ActiveHashByKey.addobject(nextnode.makeconfigkey,nextnode)
          then showmessage('Add to ActiveByKey failed #2');
          ActiveListByFVal.addobject(makefvalkey(nextnode),nextnode);
        end;
      end;


  {Make successors}
  {identify all possible moves from current position}
  {and measure distance from goal state for each}
  begin
    with  parent do
    begin
      if cols[empty]>0  {we can move empty slot left}
      then  makemove(empty,empty-1,'L');
      if rows[empty]>0  {we can move empty slot up}
      then  makemove(empty,empty-sidex,'U');
      if cols[empty]<sidex-1  {we can move empty slot right}
      then  makemove(empty,empty+1,'R');
      if rows[empty]<sidey-1 {we can move empty slot down}
      then  makemove(empty,empty+sidex,'D');
    end;
  end;



begin  {solveit}

  solving:=true;
  setlength(target,length(newtarget));
  target:=copy(newtarget,0,length(newtarget));

  try
  temp:=TMoveObj.Create(nil);
  with temp do
  begin
    {initialize the tempboard to look a lot like the real board}
    setlength(board,length(self.board));
    board:=copy(self.board,0,length(self.board));
    cumscore:=0;
    prevmove:=nil;
    empty:=self.empty;
    setscore(temp,-1,-1);
    level:=0;
    movetype:=' ';
  end;

  with statsrec do
  begin
    CountActive:=0; maxcounta:=0; maxcountChecked:=0; maxlevel:=0;
    starttime:=now;
    distance:=0;
    FVal:=0;
    count:=0; moves:=0;
    startstr:=Makeboardstr(board);
    targetstr:=Makeboardstr(target);

    path:='';
  end;
  if not ActiveHashByKey.addobject(temp.makeconfigkey,temp)
  then showmessage('Add to ActiveByKey failed #3');
  ActiveListByFval.addobject(makefvalkey(temp),temp);
  {this is the body of the search }
  result:=false;
  while (ActiveListByFVal.count>0) and (tag=0) do
  begin
    {take the best node off the open list}
    application.processmessages;
    temp:=TMoveObj(ActiveListByFVal.objects[0]);
    index:=activehashbykey.indexof(temp.makeconfigkey);

    if index<0 then
    begin
      showmessage('Potential system error, Activelistbyfval rec not found in Activelistbykey');
      exit;
    end
    else
    begin
      activeListByfval.delete(0);
      activehashbykey.delete(index);

      if temp.movescore=0 then {Solved!}
      begin
        {construct path by following the parent pointers}
        result:=true;
        solving:=false;
        statsrec.stoptime:=now;
        s:=temp.buildpath; {build the path and show it}
        statsrec.moves:=length(s);
        If assigned(OnStatsrec) then
        with statsrec do
        begin
          CountActive:=ActiveListByFVal.Count;
          Distance:=temp.movescore;
          FVAl:=temp.Fval;
          if activelistbyFVal.count>maxcounta then maxcounta:=activelistbyFVal.count;
          if checkedhash.count>maxcountchecked
          then maxcountchecked:=checkedhash.count;
          if temp.level>maxlevel then maxlevel:=temp.level;
          FollowPath(s);
          OnStatsrec(self); {update stats display}
        end;
        (*
        with activelistbyFVal do
        begin
          for i:=0 to count-1 do  TMoveObj(objects[i]).free;
        end;
        *)
        activeListByFVal.Clear; {set active count to 0 to stop loop}
      end
      else
      begin
        {find what nodes can be generated from this node}
        inc(statsrec.count);
        if (statsrec.count mod 1024 =0)and (assigned(OnStatsRec))
        then
        with statsrec do
        begin
          CountActive:=ActiveListByFVal.Count;
          distance:=temp.movescore;
          fval:=temp.fVal;
          if activelistbyFVal.count>maxcounta then maxcounta:=activelistbyFVal.count;
          if checkedhash.count>maxcountchecked
          then maxcountchecked:=checkedhash.count;
          if temp.level>maxlevel then maxlevel:=temp.level;
          stoptime:=now;
          OnStatsrec(self);
        end;
        makeSuccessors(temp);
        If not CheckedHash.addobject(temp.makeconfigkey,temp)
        then showmessage('Add to checked failed');
      end;
    end;
  end;
  finally
    ActivehashByKey.ClearAndFreeObj;

    ActiveListByFVal.clear;
    CheckedHash.clearAndFreeObj;
    solving:=false;
    //randseed:=DefaultRandomSeed;
  end;
end;


{***************** TSliderBoard.Solved *******************}
function TSliderBoard.solved:boolean;
{Check to see is board has been solved }
var
  i:integer;
begin
  if movecount=0 then result:=false
  else
  begin
    result:=true;
    for i:= low(target) to high(target) do
    begin
      if target[i]<>board[i] then
      begin
       result:=false;
       break;
      end;
    end;
  end;
end;


{********************* TSliderBoard.BeginUpdate ************}
procedure TSliderBoard.beginupdate;
{set update mode}
begin
  updatemode:=true;
end;

{******************** TSliderBoard.EndUpdate ***************}
procedure TSliderBoard.endupdate;
{reset update mode}
var
  i:integer;
begin
  updatemode:=false;
  for i:=1 to high(sliders) do sliders[i].invalidate;
end;


{*************** TSliderBoard.Parity **************}
function TSliderBoard.Parity(const testboard:TBoardRecs):integer;
{returns current board "parity", and odd or even number to
 check that it is invariate under valid moves}
var
  i,j:integer;
  count:integer;
begin
  {count inversions}
  count:=0;
  for i:=high(testboard) downto 0 do {For all cells}
  begin
    if testboard[i]=0 then   {if we have found the blank}
    begin
      if ((sidex mod 2) = 0) {and board has an even number of columns, }
      then inc(count, 1+i div sidex); {then add its row number to count}
    end
    else {otherwise we are on a numbered tile, }
    for j:=0 to i-1 do  {so for all preceding tiles,}
    if testboard[j]>testboard[i] {if it is greater than tile we are checking, }
    then inc(count); {then add 1 to count}
  end;
  result:=count mod 2;
end;


{**************** TSliderBoard.MakeBoardStr ************}
function TSliderBoard.Makeboardstr(const s:TBoardrecs):string;
{Make a string version of the board configuration}
var
  i:integer;
begin
    result:='';
    for i:= low(s) to high(s) do result:=result+','+inttostr(s[i]);
    result[1]:='(';
    result:=result+')';
 end;

{**********************************************************}
{******************** TMoveObj methods ********************}
{**********************************************************}

  {******************** TMoveObj.Create ********************}
   constructor TMoveobj.create(parentmove:TMoveObj);
   begin
     inherited create;
     if parentmove<>nil then
     begin
       {initialize the board to make it look a lot like its parent}
       setlength(board,length(parentMove.board));
       board:=copy(parentmove.board,0,length(board));
       cumscore:=parentmove.cumscore;
       movescore:=parentmove.movescore; {will be overridden later by setscore }
       fval:=0;
       prevmove:=parentmove;
       empty:=parentmove.empty;
       level:=parentmove.level+1;
     end;
   end;


{******************** TMoveObj.MakeConfigKey ***************}
    function TMoveobj.makeconfigkey:string;
    {make a kety from the board}
    var
      i:integer;
      s:string;
    begin
      s:='';
      for i:= 0 to high(board) do
      begin
        if board[i]=0 then s:=s+'E' {char($FF)}
        else s:=s+char(board[i]);
      end;
      result:=s;
    end;


{***************** TMoveObj.BuildPath ****************}
    function TMoveObj.BuildPath:string;
    {Follow moveobj path back to source, building move string as we go}
    var
      t:TMoveobj;
    begin
      result:='';
      t:=self;
      while t.prevmove<>nil do
      begin
        result:=t.movetype+result;
        t:=t.prevmove;
      end;
    end;

{**********************************************************}
{****************** Form Methods ******************}
{**********************************************************}

{****************** TForm1.FormCreate ******************}
procedure TForm1.FormCreate(Sender: TObject);
begin
  //randseed:=DefaultRandomSeed; {make results repeatable}
  SizeGrpClick(Sender);
  visited:=TStringlist.create;
  visited.sorted:=true;
  randomize;
  memo1.doublebuffered:=true;
end;

{****************** TForm1.ScrambleBtnClick ***************}
procedure TForm1.ScrambleBtnClick(Sender: TObject);
{scramble the board by moving the empty space around}
var
  n,prev,i:integer;
begin
  sizegrp.enabled:=false;
  sliderboard.beginupdate;
  i:=0;
  prev:=-1; {keep track of previous move to eliminate reversals}
  sliderboard.updatemode:=true;
  with sliderboard do
  repeat
    n:=random(4);
    case n of
      0: {left}
         if (prev<>2) and (cols[empty]>0) then
         begin
            move(empty-1);
            inc(i);
            prev:=n;
         end;
      1: {up}
         if (prev<>3) and (rows[empty]>0) then
         begin
           move(empty-sidex);
           inc(i);
           prev:=n;
         end;
      2: {right}
         if (prev<>0) and (cols[empty]<sidex-1) then
         begin
           move(empty+1);
           inc(i);
           prev:=n;
         end;
      3: {down}
         if (prev<>1) and (rows[empty]<sidey-1) then
         begin
           move(empty+sidex);
           inc(i);
           prev:=n;
         end;
    end;
  until i>=UpDown1.position;
  sliderboard.endupdate;
  sizegrp.enabled:=true;
  sliderboard.movecount:=0;
  UpdateMoveDisplay(sender);
  saveboardrecs;
  solvedpanel.visible:=false;
  showstats(sender);
end;

{************** SaveBoardrecs ***************}
procedure TForm1.SaveBoardrecs; {save the board before starting a solution}
{Called after scramble button or before starting solution search}
begin
  setlength(savedboard, length(sliderboard.board));
  with sliderboard do savedboard:=copy(board,low(board),length(board));
end;





{**************** TFom1.UpdateMoveDisplay ***************}
Procedure TForm1.UpdateMoveDisplay(sender:TObject);
{Move event exit}
begin
  form1.MoveEdt.text:=inttostr(sliderboard.movecount);
end;

{*************** TForm1.Showstats ****************}
procedure TForm1.Showstats(sender:TObject);
{Showstats event exit}
var
  s:string;
  i:integer;
  begin

    with memo1,lines, sliderboard, statsrec do
    begin
        clear;
        add('Statistics:');
        with sizegrp do add('Board Size:' + items[itemindex]);
        add('Start: '+startstr);
        add('Target:'+targetstr);
        add('Weight: ' +inttostr(weight));
        s:='';
        with heuristicsbox do
        for i:=0 to items.count-1 do if checked[i] then s:=s+items[i]+',';
        if length(s)>0 then system.delete(s,length(s),1) else s:='None';
        add('Heuristics: '+s);

        add(''); add('Debugging and progress statistics');
        add('Configs checked: '+inttostr(maxcountChecked)+', Time (Seconds): '+inttostr(trunc((stoptime-starttime)*secsperday)));
        add('Loops: ' + inttostr(count)+ ', Max.Depth: '+inttostr(maxlevel));
        add('Active Configurations: '+inttostr(countActive));
        add('Distance from solution='+inttostr(distance)+', FVal: '+inttostr(FVal));
        add('Memory in use: '+inttostr(AllocMemSize));

        add('');
        if path <>'' then add('Solution steps: '+path);
    end;
  end;

{************** TSliderboard.StartDrag *************}
procedure TSliderBoard.Startdrag(Sender: TObject; Button: TMouseButton; Shift: TShiftState;
                     X, Y: Integer);
{Mousedown exit}
{If manualmode = setup, then get ready to drag this tile}
begin
  if playmode=1 then
  with sender as TPanel do
  begin
    draggingon:=true;
    offset.x:=x;
    offset.y:=y;
    dragposition:= (top div slotw)*sidex+ left div slotw;
    with tpanel(sender) do dragstart:=point(left,top);
    bringtofront;
  end;
end;

{*************** TSliderBoard.DragIt ***********}
procedure TSliderBoard.DragIt(Sender: TObject; Shift: TShiftState; X, Y: Integer);
{Mousemove exit}
{If manual mode =setup and we are dragging then move the tile}
begin
  if (playmode=1) and (ssleft in shift) and draggingon  then
  with sender as TPanel do
  begin
    left:=left+x-offset.x;
    top:=top+y-offset.y;
  end;
end;

{********************** TSliderboard.DropIt **********8}
procedure TSliderBoard.Dropit(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
{If dragging a tile, exchange it with the drop target tile}
var
  b,b1,b2:integer;
  p1,p2:integer;
begin
  If draggingon then
  with sender as tPanel do
  begin
    left:=left+x-offset.x;
    top:=top+y-offset.y;
    {adjust to a corner position if necessary}
    if left<0 then left:=0
    else if left+width > parent.width then left:=parent.width-slotw;
    if top<0 then  top := 0
    else if top>parent.height then top:= parent.height-slotw;
    {get the tile positions in the board array}
    p1:=dragposition; {the one being dragged}
    with tpanel(sender) do p2:= ((top+y) div slotw)*sidex+ (left+x) div slotw;  {the one being dropped on}
    {get indices of the actual slider entries corresponding to these postions}
    b1:=board[p1]; b2:=board[p2];
    b:=dragstart.x;

    {if moveto slot is occupied then move it to drag tile's starting location}
    if b2>0 then
    begin
      sliders[b1].left:=sliders[b2].left;
      sliders[b2].left:=b;
      b:=dragstart.y;
      sliders[b1].top:=sliders[b2].top;
      sliders[b2].top:=b
    end
    else
    begin  {dropping on the empty slot}
      sliders[b1].left:=(p2 mod sidex)*slotw;
      sliders[b1].top:= (p2 div sidex)*slotw;
      empty:=p1;
    end;
    b:=board[p1]; board[p1]:=board[p2];  board[p2]:=b;
    with form1.memo1 do
    begin
      if (lines.count>0) and (length(lines[0])>4) and (copy(lines[0],1,4)='Stat') then
      clear;
      lines.add('Start: '+makeboardstr(board));
    end;
    draggingon:=false;
    if parity(board) = targetparity then TPanel(parent).color:=clgreen
    else TPanel(parent).color:=clred;
  end;
  form1.Saveboardrecs;
end;

{*************** TForm1.SizeGrpClick ***************}
procedure TForm1.SizeGrpClick(Sender: TObject);
{select a new board size}
var
  n:integer;
begin
  if assigned(sliderboard) then sliderboard.free;
  with sizegrp do
  begin
    n:=sizegrp.itemindex+2;
    case sizegrp.itemindex of
    0..3:
      begin
        sidex:=n;
        sidey:=n
      end;
    4:
      begin
        sidex:=4; sidey:=2;
      end;
    5:
      begin
        sidex:=4; sidey:=3;
      end;
    end; {case}
  end;
  Sliderboard:=TSliderBoard.create(panel1,sidex,sidey);
  solvedpanel.visible:=false;
  with sliderboard do
  begin
    OnMove:=UpdateMoveDisplay;
    OnStatsRec:=ShowStats;
    TargetRGrpClick(sender);
    PlayModeGrpClick(sender);
    Memo1.clear;
    memo1.lines.add('Parity for current boardsize and target is '+inttostr(targetparity));
  end;
end;

{********************* TFrom1.SolveBtnClick **************}
procedure TForm1.SolveBtnClick(sender:TObject);
{Start autosolve action}
var
  target:TBoardrecs;
  i:byte;
  temp:TMoveObj;
  OK:boolean;
begin

  with sliderboard do
  begin
    if parity(board)<>targetparity then
    begin
      showmessage('Parities differ, no soluton possible');
      exit;
    end;
    weight:=strtoint(weightedt.text);
    if weight=0 then begin weight:=1;  weightedt.text:='1'; end;
    tag:=0;
    movecount:=0;
    replayindex:=0;
  end;
  cursor:=crhourglass;
  cancelbtn.visible:=true;
  Solvedpanel.visible:=false;
  btngroup.enabled:=false;
  solvebtn.enabled:=false;

  try
    OK:=true;
    with sliderboard do
    for i:=0 to high(board) do if board[i]<>target[i] then
    begin
      OK:=false;
      break;
    end;
    if OK then
    begin
      showmessage('Board already solved, no action taken');
      exit;
    end;
    saveboardrecs;
    target:=Sliderboard.MakeTargetBoard1(i);
    {Try to solve the board!}
    If not sliderboard.solveit(target) then showmessage('No solution found');

  finally
    cancelbtn.visible:=false;
    //startoverbtn.visible:=true;
    SolvedPanel.visible:=true;
    cursor:=crdefault;
    btngroup.enabled:=true;
    solvebtn.enabled:=true;
    with sliderboard do
    begin
      (*
      Activehashbykey.freeandclearobjects;

      ActiveHashByKey:=THashStr.create(99991,0.50,false{true=keepstats});
      Checkedhash.freeandclearobjects;;
      CheckedHash:=THashStr.create(99991,0.50,false);
      ActivelistByFval.free;
      ActiveListByFVal:=TStringList.create;
      ActivelistByFval.sorted:=true;
      *)
    end;
     showstats(sender);
  end;
end;

{***************** TForm1.CancelbtnClick *************}
procedure TForm1.CancelBtnClick(Sender: TObject);
{Stop solving}
begin
  Sliderboard.tag:=1;
end;

{*************** Tform1.ResetBtnClick *************}
procedure TForm1.ResetBtnClick(Sender: TObject);
{Reset the board}
begin
  sliderboard.resetboard;
  cheatcount:=0;
  pCount:=0;  {parity check count}
  solvedpanel.visible:=false;
  sliderboard.playmode:=playmodegrp.itemindex;
  showstats(sender);
end;

{**************** HeuristicsBoxClick ***********}
procedure TForm1.HeuristicsBoxClick(Sender: TObject);
{Set additional heuristics options}
begin
  with HeuristicsBox
  do sliderboard.setheuristics(checked[0],checked[1],checked[2]);
end;

{************** TargetRGrpClick *************}
procedure TForm1.TargetRgrpClick(Sender: TObject);
{User selected a new target pattern}
begin
  with sliderboard do
  begin
    targettype:=TargetRGrp.itemindex;
    resetboard;
    memo1.clear;
    memo1.lines.add('Parity for current boardsize and target is '+inttostr(targetparity));
    solvedpanel.visible:=false;
  end;
end;

{************** FormKeyPress ****************}
procedure TForm1.FormKeyPress(Sender: TObject; var Key: Char);
{Check for P (Parity) or C (Cheat- exchange two random adjacent tiles)}
var
  b:byte;
  x,y:integer;
  OK:Boolean;
  p1,p2:integer;
begin
  if (pcount=0) and (cheatcount=0) then memo1.clear;
  if upcase(key)='P' then
  with memo1 do
  begin
    inc(Pcount);
    lines.add('Parity='+inttostr(sliderboard.parity(sliderboard.board)));
  end;
  if upcase(key)='C' then
  with memo1, sliderboard  do
  begin
    inc(cheatcount);
    lines.add('Cheat count '+ inttostr(cheatcount));

    OK:=false;
    repeat
      p1:=random(sidex*sidey-1)+1;
      case random(3) of
        0: {right}p2:=p1+1;
        1: {left}p2:=p1-1;
        2: {up}p2:=p1-sidex;
        3: {down}p2:=p1+sidex;
        else p2:=0; {can't happen, added just to get rid of compiler warning message}
      end;
      if (p2>=1) and (p2<=sidex*sidey) and (board[p1-1]<>0) and (board[p2-1]<>0)
      then OK:=true;
    until OK;
    x:=board[p1-1]; y:=board[p2-1];
    b:=sliders[x].left; sliders[x].left:=sliders[y].left; sliders[y].left:=b;
    b:=sliders[x].top; sliders[x].top:=sliders[y].top; sliders[y].top:=b;
    b:=board[p1-1]; board[p1-1]:=board[p2-1];  board[p2-1]:=b;
    if parity(board) <> targetparity
    then sliderboard.color:=clred else sliderboard.color:=clgreen;
    lines.add(format('Cheat count %d,  Parity=%d',[cheatcount,sliderboard.parity(board)]));
    invalidate;
  end;
end;


{*************** StartOverBtnClick ***********}
procedure TForm1.StartOverBtnClick(Sender: TObject);
var
  i,n:integer;
begin
  with sliderboard do
  begin
    endupdate;
    sizegrp.enabled:=true;
    playmode:=playmodegrp.itemindex;
    if length(savedboard)>0 then
    begin
      board:=copy(savedboard,low(savedboard),length(savedboard));
      for i:=0 to sidex*sidey-1 do
      begin
        n:=board[i];
        if n>0
        then
        with sliders[n] do
        begin
          left:= cols[i]*slotw;
          top:=  rows[i]*slotw;
        end
        else empty:=i;
      end;
    end;
    movecount:=0;
    replayindex:=0;
  end;
  UpdateMoveDisplay(sender);
  showstats(sender);
end;




{*************** PlayModeGrpClick ************}
procedure TForm1.PlayModeGrpClick(Sender: TObject);
begin
  sliderboard.playmode:=playmodegrp.itemindex;
  if playmodegrp.itemindex=1 then memo1.lines.clear;
end;

{************* ClearStatsBtnClick ************}
procedure TForm1.ClearStatsBtnClick(Sender: TObject);
begin
   memo1.Clear;
   curstatstart:=0;
   curstatsize:=0;
end;

{************* ReplayBtnClick **********}
procedure TForm1.ReplayBtnClick(Sender: TObject);
begin
  with sliderboard do
  begin
    If (replayindex=0)  then
    begin
      startoverbtnclick(sender);
      application.processmessages;
      sleep(500);
      replayindex:=1;
      playmode:=2;
      clicksOK:=false;
    end;
    move(replaymoves[replayindex]);
    if Solved then  showmessage('Replay complete');
    replayindex:=(replayindex+1)mod length(replaymoves); {wrap to 0 after last move}
    if replayindex=0 then playmode:=playmodegrp.itemindex; {OK to restore noremal playmode}
  end;
end;

procedure TForm1.StaticText1Click(Sender: TObject);
begin
   ShellExecute(Handle, 'open', 'http://www.delphiforfun.org/',
  nil, nil, SW_SHOWNORMAL) ;
end;

end.


